package at.ac.uibk.gitsearch.config;

import static org.springframework.http.MediaType.APPLICATION_FORM_URLENCODED_VALUE;

import java.net.URI;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import javax.servlet.DispatcherType;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.web.servlet.FilterRegistrationBean;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Import;
import org.springframework.core.Ordered;
import org.springframework.core.convert.converter.Converter;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.RequestEntity;
import org.springframework.security.authentication.AbstractAuthenticationToken;
import org.springframework.security.authentication.AuthenticationProvider;
import org.springframework.security.config.annotation.method.configuration.EnableGlobalMethodSecurity;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.builders.WebSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.security.oauth2.client.endpoint.DefaultAuthorizationCodeTokenResponseClient;
import org.springframework.security.oauth2.client.endpoint.OAuth2AccessTokenResponseClient;
import org.springframework.security.oauth2.client.endpoint.OAuth2AuthorizationCodeGrantRequest;
import org.springframework.security.oauth2.client.endpoint.OAuth2AuthorizationCodeGrantRequestEntityConverter;
import org.springframework.security.oauth2.client.oidc.userinfo.OidcUserRequest;
import org.springframework.security.oauth2.client.oidc.userinfo.OidcUserService;
import org.springframework.security.oauth2.client.registration.ClientRegistration;
import org.springframework.security.oauth2.client.userinfo.OAuth2UserService;
import org.springframework.security.oauth2.core.OAuth2TokenValidator;
import org.springframework.security.oauth2.core.endpoint.OAuth2AuthorizationExchange;
import org.springframework.security.oauth2.core.endpoint.OAuth2ParameterNames;
import org.springframework.security.oauth2.core.endpoint.PkceParameterNames;
import org.springframework.security.oauth2.core.oidc.OidcUserInfo;
import org.springframework.security.oauth2.core.oidc.user.DefaultOidcUser;
import org.springframework.security.oauth2.core.oidc.user.OidcUser;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.jwt.JwtDecoder;
import org.springframework.security.oauth2.jwt.JwtDecoders;
import org.springframework.security.oauth2.jwt.JwtValidators;
import org.springframework.security.oauth2.jwt.NimbusJwtDecoder;
import org.springframework.security.oauth2.server.resource.BearerTokenAuthenticationToken;
import org.springframework.security.oauth2.server.resource.authentication.JwtAuthenticationConverter;
import org.springframework.security.web.authentication.AuthenticationSuccessHandler;
import org.springframework.security.web.csrf.CookieCsrfTokenRepository;
import org.springframework.security.web.csrf.CsrfFilter;
import org.springframework.security.web.header.writers.ReferrerPolicyHeaderWriter;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.filter.CorsFilter;
import org.springframework.web.filter.ForwardedHeaderFilter;
import org.springframework.web.util.UriComponentsBuilder;
import org.zalando.problem.spring.web.advice.security.SecurityProblemSupport;

import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.security.jwt.JWTConfigurer;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.security.oauth2.SavedRequestAwareAuthenticationSuccessHandlerWithJWTSupport;
import at.ac.uibk.gitsearch.security.oauth2.UserDetailsFetcher;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.dto.UserDTO;
import at.ac.uibk.gitsearch.service.mapper.UserMapper;

@EnableWebSecurity
@EnableGlobalMethodSecurity(prePostEnabled = true, securedEnabled = true)
@Import(SecurityProblemSupport.class)
public class SecurityConfiguration extends WebSecurityConfigurerAdapter {

    @Value("${spring.security.oauth2.client.provider.gitlabOidc.issuer-uri}")
    private String gitLabIssuerUri;
    
    private final TokenProvider tokenProvider;

    private final CorsFilter corsFilter;
    private final SecurityProblemSupport problemSupport;
	private UserDetailsFetcher userDetailsFetcher;


    public SecurityConfiguration(TokenProvider tokenProvider, CorsFilter corsFilter, 
    		UserDetailsFetcher userDetailsFetcher,
    		SecurityProblemSupport problemSupport
    		) {
        this.tokenProvider = tokenProvider;
        this.corsFilter = corsFilter;
        this.problemSupport = problemSupport;
        this.userDetailsFetcher = userDetailsFetcher;
    }

    @Bean
    public PasswordEncoder passwordEncoder() {
        return new BCryptPasswordEncoder();
    }

    @Override
    public void configure(WebSecurity web) {
        web.ignoring()
            .antMatchers(HttpMethod.OPTIONS, "/**")
            .antMatchers("/app/**/*.{js,html}")
            .antMatchers("/i18n/**")
            .antMatchers("/content/**")
            .antMatchers("/h2-console/**")
            .antMatchers("/swagger-ui/index.html")
            .antMatchers("/test/**");
    }

    @Override
    public void configure(HttpSecurity http) throws Exception {
        // @formatter:off
        http
        .csrf()
        .csrfTokenRepository(CookieCsrfTokenRepository.withHttpOnlyFalse())
        .ignoringAntMatchers("/api/pluginIF/**") // Allowing plugin Access. security is checked by token
	    .and()
	    .authenticationProvider(customJWTauthenticationProvider())
	    .addFilterBefore(corsFilter, CsrfFilter.class)
//            .addFilterBefore(corsFilter, UsernamePasswordAuthenticationFilter.class)
            .exceptionHandling()
                .authenticationEntryPoint(problemSupport)
                .accessDeniedHandler(problemSupport)
        .and()
        .addFilterBefore(forwardedHeaderFilter().getFilter(), CsrfFilter.class)
        .headers()
        	// allow images from everywhere
        	.contentSecurityPolicy("default-src 'self'; frame-src 'self' data:; script-src 'self' 'unsafe-inline' 'unsafe-eval' https://storage.googleapis.com; style-src 'self' 'unsafe-inline'; img-src *; font-src 'self' data:")
        .and()
            .referrerPolicy(ReferrerPolicyHeaderWriter.ReferrerPolicy.STRICT_ORIGIN_WHEN_CROSS_ORIGIN)
        .and()
            .featurePolicy("geolocation 'none'; midi 'none'; sync-xhr 'none'; microphone 'none'; camera 'none'; magnetometer 'none'; gyroscope 'none'; speaker 'none'; fullscreen 'self'; payment 'none'")
        .and()
            .frameOptions()
            .deny()
        .and()
            .sessionManagement()
            .sessionCreationPolicy(SessionCreationPolicy.STATELESS)
        .and()
            .authorizeRequests()
            .antMatchers("/api/statistics/**").permitAll()
            .antMatchers("/oauth2/**").permitAll()
            .antMatchers("/api/search/**").permitAll() // search is always allowed, may return more data, if authenticated
            .antMatchers("/api/pluginIF/**").permitAll() // plugins calls are always allowed, security by tokens
            .antMatchers("/api/applicationInfo/**").permitAll() // everybody may retrieve deployment infos
            .antMatchers("/api/authenticate").permitAll()
            .antMatchers("/api/exerciseFile/**").permitAll()
            .antMatchers("/api/exercise/**").permitAll()
            .antMatchers("/api/refreshToken").permitAll()
            .antMatchers("/api/register").denyAll()
            .antMatchers("/api/activate").permitAll()
            .antMatchers("/api/account/reset-password/init").permitAll()
            .antMatchers("/api/account/reset-password/finish").permitAll()
            .antMatchers("/api/numberOfLikes/**").permitAll()
            .antMatchers("/api/hasLiked/**").permitAll()
            .antMatchers("/api/**").authenticated()
            .antMatchers("/management/health").permitAll()
            .antMatchers("/management/info").permitAll()
            .antMatchers("/management/prometheus").permitAll()
            .antMatchers("/management/**").hasAuthority(AuthoritiesConstants.ADMIN)
            .antMatchers("/api/currentuser-watch-lists/**").authenticated()
        .and()
            .httpBasic()
        .and()
        .apply(securityConfigurerAdapter())
        .and()
        	.oauth2Login(oauth2 -> oauth2
                .userInfoEndpoint(userInfo -> userInfo
                    .oidcUserService(this.oidcUserService())))
            .oauth2Login().successHandler(getAuthenticationSuccessHandler())
        .and()
            .oauth2ResourceServer() // .bearerTokenResolver(bearerTokenResolver)
                .jwt()
                  .jwtAuthenticationConverter(authenticationConverter())
                
                .and()
//                .authenticationManagerResolver(authenticationManagerResolver())
            .and()
                .oauth2Client();

        http.oauth2Login().tokenEndpoint().accessTokenResponseClient(accessTokenResponseClient());

        	
//            .apply(securityConfigurerAdapter())  
              // @formatter:on
    }

    Converter<Jwt, AbstractAuthenticationToken> authenticationConverter() {
        JwtAuthenticationConverter jwtAuthenticationConverter = new JwtAuthenticationConverter();
        jwtAuthenticationConverter.setJwtGrantedAuthoritiesConverter(new at.ac.uibk.gitsearch.security.oauth2.JwtGrantedAuthorityConverter());
        return jwtAuthenticationConverter;
    }
    
    private JWTConfigurer securityConfigurerAdapter() {
        return new JWTConfigurer(tokenProvider);
    }


    public AuthenticationProvider customJWTauthenticationProvider() {
    	return new AuthenticationProvider() {
			
			@Override
			public boolean supports(Class<?> authentication) {
				return BearerTokenAuthenticationToken.class.isAssignableFrom(authentication);
			}
			
			@Override
			public Authentication authenticate(Authentication authentication) throws AuthenticationException {
				if(authentication.isAuthenticated()) return authentication;
				if (authentication instanceof BearerTokenAuthenticationToken) {
					BearerTokenAuthenticationToken bearerToken = (BearerTokenAuthenticationToken) authentication;
					if(tokenProvider.validateToken(bearerToken.getToken())) {
						return tokenProvider.getAuthentication(bearerToken.getToken());
					}
					
				}
				return null;
			}
		};
    }
    
    /**
     * not used by OidcAuthorizationCodeAuthenticationProvider :-(
     * @return
     */
    @Bean
    JwtDecoder jwtDecoder() {
        NimbusJwtDecoder oidcJwtDecoder = (NimbusJwtDecoder) JwtDecoders.fromOidcIssuerLocation(gitLabIssuerUri);

        OAuth2TokenValidator<Jwt> withIssuer = JwtValidators.createDefaultWithIssuer(gitLabIssuerUri);

        oidcJwtDecoder.setJwtValidator(withIssuer);
        
        return oidcJwtDecoder;
    }
    
    // @Bean
    private OAuth2AccessTokenResponseClient<OAuth2AuthorizationCodeGrantRequest> accessTokenResponseClient(){
        DefaultAuthorizationCodeTokenResponseClient accessTokenResponseClient = 
          new DefaultAuthorizationCodeTokenResponseClient(); 
        accessTokenResponseClient.setRequestEntityConverter(new CustomOAuth2AuthorizationCodeGrantRequestEntityConverter()); 
        return accessTokenResponseClient;
    }
    
    
//    @Bean
    public FilterRegistrationBean<ForwardedHeaderFilter> forwardedHeaderFilter() {
        ForwardedHeaderFilter filter = new ForwardedHeaderFilter();
        FilterRegistrationBean<ForwardedHeaderFilter> registration = new FilterRegistrationBean<>(filter);
        registration.setDispatcherTypes(DispatcherType.REQUEST, DispatcherType.ASYNC, DispatcherType.ERROR);
        registration.setOrder(Ordered.HIGHEST_PRECEDENCE);
        registration.setUrlPatterns(Collections.singletonList("/oauth2"));
        return registration;
    }

    /**
     * corrects UTF-8 Handling.
     * @author Michael Breu
     *
     */
    public static class CustomOAuth2AuthorizationCodeGrantRequestEntityConverter extends OAuth2AuthorizationCodeGrantRequestEntityConverter {
    
    	/**
    	 * Returns the {@link RequestEntity} used for the Access Token Request.
    	 *
    	 * @param authorizationCodeGrantRequest the authorization code grant request
    	 * @return the {@link RequestEntity} used for the Access Token Request
    	 */
    	@Override
    	public RequestEntity<?> convert(OAuth2AuthorizationCodeGrantRequest authorizationCodeGrantRequest) {
    		ClientRegistration clientRegistration = authorizationCodeGrantRequest.getClientRegistration();

    		HttpHeaders headers = getTokenRequestHeaders(clientRegistration);
    		MultiValueMap<String, String> formParameters = this.buildFormParameters(authorizationCodeGrantRequest);
    		URI uri = UriComponentsBuilder.fromUriString(clientRegistration.getProviderDetails().getTokenUri())
    				.build()
    				.toUri();

    		return new RequestEntity<>(formParameters, headers, HttpMethod.POST, uri);
    	}

    	static HttpHeaders getTokenRequestHeaders(ClientRegistration clientRegistration) {
    		HttpHeaders headers = new HttpHeaders();
    		headers.addAll(getDefaultTokenRequestHeaders());
//    		if (ClientAuthenticationMethod.BASIC.equals(clientRegistration.getClientAuthenticationMethod())) {
//    			headers.setBasicAuth(clientRegistration.getClientId(), clientRegistration.getClientSecret());
//    		}
    		return headers;
    	}
    	
    	private static HttpHeaders getDefaultTokenRequestHeaders() {
    		HttpHeaders headers = new HttpHeaders();
    		headers.setAccept(Collections.singletonList(MediaType.APPLICATION_JSON));
    		final MediaType contentType = MediaType.valueOf(APPLICATION_FORM_URLENCODED_VALUE + ";charset=UTF-8");
    		headers.setContentType(contentType);
    		return headers;
    	}

    	/**
    	 * Returns a {@link MultiValueMap} of the form parameters used for the Access Token Request body.
    	 *
    	 * @param authorizationCodeGrantRequest the authorization code grant request
    	 * @return a {@link MultiValueMap} of the form parameters used for the Access Token Request body
    	 */
    	private MultiValueMap<String, String> buildFormParameters(OAuth2AuthorizationCodeGrantRequest authorizationCodeGrantRequest) {
    		ClientRegistration clientRegistration = authorizationCodeGrantRequest.getClientRegistration();
    		OAuth2AuthorizationExchange authorizationExchange = authorizationCodeGrantRequest.getAuthorizationExchange();

    		MultiValueMap<String, String> formParameters = new LinkedMultiValueMap<>();
    		formParameters.add(OAuth2ParameterNames.GRANT_TYPE, authorizationCodeGrantRequest.getGrantType().getValue());
    		formParameters.add(OAuth2ParameterNames.CODE, authorizationExchange.getAuthorizationResponse().getCode());
    		String redirectUri = authorizationExchange.getAuthorizationRequest().getRedirectUri();
    		String codeVerifier = authorizationExchange.getAuthorizationRequest().getAttribute(PkceParameterNames.CODE_VERIFIER);
    		if (redirectUri != null) {
    			formParameters.add(OAuth2ParameterNames.REDIRECT_URI, redirectUri);
    		}

			formParameters.add(OAuth2ParameterNames.CLIENT_ID, clientRegistration.getClientId());

//    		if (!ClientAuthenticationMethod.BASIC.equals(clientRegistration.getClientAuthenticationMethod())) {
//    			formParameters.add(OAuth2ParameterNames.CLIENT_ID, clientRegistration.getClientId());
//    		}
//    		if (ClientAuthenticationMethod.POST.equals(clientRegistration.getClientAuthenticationMethod())) {
//    			formParameters.add(OAuth2ParameterNames.CLIENT_SECRET, clientRegistration.getClientSecret());
//    		}
    		if (codeVerifier != null) {
    			formParameters.add(PkceParameterNames.CODE_VERIFIER, codeVerifier);
    		}

    		return formParameters;
    	}
    

    }
    
   public AuthenticationSuccessHandler getAuthenticationSuccessHandler() {
	   return new SavedRequestAwareAuthenticationSuccessHandlerWithJWTSupport(tokenProvider);
   }
   
   private OAuth2UserService<OidcUserRequest, OidcUser> oidcUserService() {
       final OidcUserService delegate = new OidcUserService();

       final Set<String> defaultAuthorities = new HashSet<>();
       defaultAuthorities.add(AuthoritiesConstants.USER);
       final UserService userService = getApplicationContext().getBean(UserService.class);
       final UserMapper userMapper = getApplicationContext().getBean(UserMapper.class);

       return userRequest -> {
           // Delegate to the default implementation for loading a user
           OidcUser oidcUser = delegate.loadUser(userRequest);

           Set<GrantedAuthority> mappedAuthorities = new HashSet<>();
           mappedAuthorities.addAll(oidcUser.getAuthorities());
    	
        	String email = oidcUser.getEmail();
			
	        Optional<User> userO = userService.getUserWithAuthoritiesByEmail(email);
	        
	        if(userO.isPresent()) {
	        	UserDTO userDto = userMapper.userToUserDTO(userO.get());
	        	boolean hasChanged = userDetailsFetcher.updateUserDetails(userDto, oidcUser, userRequest.getAccessToken(), userRequest.getClientRegistration());
	        	if (hasChanged)
	        		userService.updateUser(userDto);
	        	
	        	userO.get().getAuthorities()
	        	  .forEach(
	        			  
	        			auth -> { 
	        				if(mappedAuthorities.stream()
	        			  .filter(existingAuth -> existingAuth.getAuthority().equals(auth.getName()))
	        			  .findAny().isEmpty()) {
	        					mappedAuthorities.add(new SimpleGrantedAuthority(auth.getName()));}
	        				}
	        			);
	        } else {
	        	UserDTO u = new UserDTO();
	        	u.setFirstName(oidcUser.getGivenName());
	        	u.setLastName(oidcUser.getFamilyName());
	        	u.setLogin(email);
	        	u.setActivated(true);
	        	u.setLastModifiedBy("system");
	        	u.setCreatedDate(java.time.Instant.now());

	        	u.setAuthorities(defaultAuthorities);
	        	u.setEmail(oidcUser.getEmail());
	        	@SuppressWarnings("unused")
				boolean hasChanged = userDetailsFetcher.updateUserDetails(u, oidcUser, userRequest.getAccessToken(), userRequest.getClientRegistration());
	        	userService.createUser(u);
	        }

	       Map<String, Object> claims = new HashMap<>( oidcUser.getUserInfo().getClaims());
	       claims.put(TokenProvider.GITLAB_ACCESS_TOKEN, userRequest.getAccessToken().getTokenValue());
	       claims.put(TokenProvider.GITLAB_ACCESS_ISSUER, userRequest.getClientRegistration().getRegistrationId());
	       OidcUserInfo userInfo = new OidcUserInfo(claims);
	           
           oidcUser = new DefaultOidcUser(mappedAuthorities, oidcUser.getIdToken(), userInfo);

           return oidcUser;
       };
   }
    
}

