package at.ac.uibk.gitsearch.service;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.ProjectApi;
import org.gitlab4j.api.models.Group;
import org.gitlab4j.api.models.RepositoryFile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;
import org.springframework.util.StringUtils;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;

/**
 * Service for accessing git lab backbone
 * <p>
 */
@Service
public class GitlabService {
	
	@Autowired
	private GitLabRepository gitLabRepository;
	@Autowired
	private MetaDataRepository metaDataRepository;
	@Autowired
	protected TokenProvider tokenProvider;
	
	private final Logger log = LoggerFactory.getLogger(ShoppingBasketService.class);


    public Boolean repositoryExists(String projectID) {
		try{
			final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
			final ProjectApi gitLabProjectApi = gitLabApi.getProjectApi();
			return gitLabProjectApi.getProject(projectID) != null;}
		catch(GitLabApiException e){
			log.error(e.getMessage(), e);
			return false;
		}
    }

    /**
     * returns an authentication of the current user against gitlab, together with groups and access token
     */
    public Authentication getGitLabAuthentication(String name, String oAuth2Token) {
    	final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(oAuth2Token);
    	List<Group> groups = Collections.emptyList();
    	if(!org.apache.commons.lang3.StringUtils.isEmpty(oAuth2Token))
		  try {
				groups = gitLabApi.getGroupApi().getGroups();
			  } catch (GitLabApiException e) {
				log.warn("Cannot read groups for {}", name);
				groups = Collections.emptyList();
			  }
    	
    	
        Collection<? extends GrantedAuthority> authorities =
                groups.stream().map(g -> g.getFullPath())
                    .map(SimpleGrantedAuthority::new)
                    .collect(Collectors.toList());

        User principal = new User(name,"", authorities);

        final UsernamePasswordAuthenticationToken authentication = new UsernamePasswordAuthenticationToken(principal, "", authorities);
    	Map<String, String> detailsMap = new HashMap<>();
      		detailsMap.put(TokenProvider.GITLAB_ACCESS_TOKEN, oAuth2Token);
  			detailsMap.put(TokenProvider.GITLAB_ACCESS_ISSUER, "who cares?");
       	authentication.setDetails(detailsMap);
		return authentication;
    	
    }
    
	
	public InputStream getRepositoryZip(String exerciseID) throws GitLabApiException, IOException {
		
        final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
		return rePackageGitLabProjectZip(new ZipInputStream(gitLabApi.getRepositoryApi().getRepositoryArchive(exerciseID, "HEAD", "zip")), "from project " + exerciseID);
	}
	
	/**
	 * chops of the main folder, and brings every file one level up. Also deleting
	 * all plain files in main folder.
	 * 
	 * @param zipIn the zip to be repackaged
	 * @param originalLocation the original location. For debug purposes only.
	 * @return
	 * @throws IOException
	 */
	protected InputStream rePackageGitLabProjectZip(ZipInputStream zipIn, String originalLocation) throws IOException {
		final PipedInputStream pis = new PipedInputStream(1024*256);
		final PipedOutputStream pos = new PipedOutputStream(pis);

		Runnable rePackage = () -> {
			try (ZipOutputStream zipOut = new ZipOutputStream(pos)) {

				ZipEntry zipInEntry = zipIn.getNextEntry();
				String prefix = "";
				if (zipInEntry.isDirectory()) { // main directory is prefix to all entries
					prefix = zipInEntry.getName();
				}

				while (zipInEntry != null) {
					if (zipInEntry.getName().startsWith(prefix)) {
						String newName = zipInEntry.getName().substring(prefix.length());
						if (!StringUtils.isEmpty(newName)) {
							ZipEntry zipOutEntry = new ZipEntry(newName);
							if (zipInEntry.isDirectory()) {
							log.debug("ignoring directory {}", zipInEntry.getName());
								zipOut.putNextEntry(zipOutEntry);
							} else {
								zipOut.putNextEntry(zipOutEntry);
								StreamUtils.copy(zipIn, zipOut);
								zipOut.closeEntry();
							}
						}}
					zipInEntry = zipIn.getNextEntry();
				}
				zipIn.close();
			} catch (IOException e) {
				log.error("Cannot rezip file from {}", originalLocation);
			}
		};

		new Thread(rePackage).start();

		return pis;

	}
	
	public InputStream getRepositoryFile(ExerciseId exerciseId, String filePath)
			throws GitLabApiException, IOException {

		final SearchResultDTO exercise = metaDataRepository.getExerciseById(exerciseId);
		final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
		RepositoryFile file;
		file = gitLabApi.getRepositoryFileApi().getFile(exercise.getProject().getProjectId(),
				exerciseId.extendPath(filePath), "master");
		return new ByteArrayInputStream(file.getDecodedContentAsBytes());
	}
	

}
