package at.ac.uibk.gitsearch.web.rest;

import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.util.Optional;

import javax.servlet.http.HttpServletRequest;

import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.AntPathMatcher;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.HandlerMapping;

import at.ac.uibk.gitsearch.service.GitlabService;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.web.util.HeaderUtil;

/**
 * this resource should give access to various aspects of an exercise or course
 * @author Michael Breu
 *
 */
@RestController
@RequestMapping("/api")
public class ExerciseResource {
	
    private static final String NOT_FOUND = "not found";

	private static final String REPOSITORYFILE = "RepositoryFile";

	private final Logger log = LoggerFactory.getLogger(ExerciseResource.class);
    
    @Autowired
    private GitlabService gitLabService;
    @Autowired
    private SearchService searchService;
    
    @Value("${jhipster.clientApp.name}")
    private String applicationName;


    /**
     * GET some file of an exercise.
     *
     * @param filePath the file path
     * @return the ResponseEntity with status 200 (OK) and with body the exercise file, or with status 404 (Not Found)
     */
    @GetMapping("/exerciseFile/**")
    public ResponseEntity<Resource> getRepositoryFile(HttpServletRequest request, @RequestParam("filePath") String filePath) throws IOException {
		String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

		String exerciseId = new AntPathMatcher().extractPathWithinPattern(pattern, request.getRequestURI());

		ExerciseId parsedId;
		try {
			parsedId = ExerciseId.fromString(exerciseId);
		} catch (ParseException e) {
            return ResponseEntity.notFound().headers(HeaderUtil.createFailureAlert(applicationName, true, REPOSITORYFILE, NOT_FOUND,
                    "There was an error finding your exercise " + exerciseId + ".")).build();        
		}

        InputStream repositoryStream = null;
        try {
            repositoryStream = gitLabService.getRepositoryFile(parsedId, filePath);
        } catch (IOException | GitLabApiException e) {
            log.warn("Cannot read file for exercise {} and path {}: {}", exerciseId, filePath, e.getMessage());
            return ResponseEntity.notFound().headers(HeaderUtil.createFailureAlert(applicationName, true, REPOSITORYFILE, NOT_FOUND,
                    "There was an error on the server and the resource file could not be created.")).build();        }

        InputStreamResource resource = new InputStreamResource(repositoryStream);

        return ResponseEntity.ok().contentType(MediaType.APPLICATION_OCTET_STREAM).header("filename", filePath).body(resource);
    }

    /**
     * GET the exercise by Id
     * 
     * @param request the complete request, in order to parse the exerciseId
     * @return the ResponseEntity with status 200 (OK) and with the searchresult, or with status 404 (Not Found)
     */
    @GetMapping("/exercise/**")
    public ResponseEntity<SearchResultDTO> getExerciseById(HttpServletRequest request) throws IOException {
		String pattern = (String) request.getAttribute(HandlerMapping.BEST_MATCHING_PATTERN_ATTRIBUTE);

		String exerciseId = new AntPathMatcher().extractPathWithinPattern(pattern, request.getRequestURI());

		ExerciseId parsedId;
		try {
			parsedId = ExerciseId.fromString(exerciseId);
		} catch (ParseException e) {
            return ResponseEntity.notFound().headers(HeaderUtil.createFailureAlert(applicationName, true, REPOSITORYFILE, NOT_FOUND,
                    "There was an error with your requested id " + exerciseId + ".")).build();        
		}

		final Optional<SearchResultDTO> result = searchService.findExerciseById(parsedId);
		

		if(result.isPresent()) {
			return ResponseEntity.ok().contentType(MediaType.APPLICATION_JSON).body(result.get());
		} else {
            return ResponseEntity.notFound().headers(HeaderUtil.createFailureAlert(applicationName, true, REPOSITORYFILE, NOT_FOUND,
                    "There was an error finding your exercise " + exerciseId + ".")).build();        
		}
    }

}
