package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.domain.Likes;
import at.ac.uibk.gitsearch.service.LikesService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;
import at.ac.uibk.gitsearch.service.dto.LikesCriteria;
import at.ac.uibk.gitsearch.service.LikesQueryService;

import io.github.jhipster.web.util.HeaderUtil;
import io.github.jhipster.web.util.ResponseUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.net.URI;
import java.net.URISyntaxException;
import java.time.LocalDate;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.stream.StreamSupport;

import static org.elasticsearch.index.query.QueryBuilders.*;

/**
 * REST controller for managing {@link at.ac.uibk.gitsearch.domain.Likes}.
 */
@RestController
@RequestMapping("/api")
public class LikesResource {

    private final Logger log = LoggerFactory.getLogger(LikesResource.class);

    private static final String ENTITY_NAME = "likes";

    @Value("${jhipster.clientApp.name}")
    private String applicationName;

    private final LikesService likesService;
    private final UserService userService;

    private final LikesQueryService likesQueryService;

    public LikesResource(LikesService likesService, LikesQueryService likesQueryService, UserService userService) {
        this.likesService = likesService;
        this.likesQueryService = likesQueryService;
        this.userService = userService;
    }

    /**
     * {@code POST  /likes} : Create a new likes.
     *
     * @param likes the likes to create.
     * @return the {@link ResponseEntity} with status {@code 201 (Created)} and with
     *         body the new likes, or with status {@code 400 (Bad Request)} if the
     *         likes has already an ID.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PostMapping("/likes")
    public ResponseEntity<Likes> createLikes(@Valid @RequestBody Likes likes) throws URISyntaxException {
        log.debug("REST request to save Likes : {}", likes);
        if (likes.getId() != null) {
            throw new BadRequestAlertException("A new likes cannot already have an ID", ENTITY_NAME, "idexists");
        }
        Likes result = likesService.save(likes);
        return ResponseEntity
                .created(new URI("/api/likes/" + result.getId())).headers(HeaderUtil
                        .createEntityCreationAlert(applicationName, true, ENTITY_NAME, result.getId().toString()))
                .body(result);
    }

    /**
     * {@code PUT  /likes} : Updates an existing likes.
     *
     * @param likes the likes to update.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body
     *         the updated likes, or with status {@code 400 (Bad Request)} if the
     *         likes is not valid, or with status
     *         {@code 500 (Internal Server Error)} if the likes couldn't be updated.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PutMapping("/likes")
    public ResponseEntity<Likes> updateLikes(@Valid @RequestBody Likes likes) throws URISyntaxException {
        log.debug("REST request to update Likes : {}", likes);
        if (likes.getId() == null) {
            throw new BadRequestAlertException("Invalid id", ENTITY_NAME, "idnull");
        }
        Likes result = likesService.save(likes);
        return ResponseEntity.ok().headers(
                HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, likes.getId().toString()))
                .body(result);
    }

    @PutMapping("/likeProject")
    public ResponseEntity<Likes> updateLikeWithProjectID(@RequestBody Likes likes) {

        if (likes.getProjectID() != null) {
            Likes like = likesService.findLikesByUserIDandProjectID(userService.getUserWithAuthorities().get().getId().intValue(), likes.getProjectID());
            
            if(like == null){
                log.debug("Like not known atm creating new one");
                like = new Likes();
                like.setUserID(userService.getUserWithAuthorities().get().getId().intValue());
                like.setProjectID(likes.getProjectID());
            }
                like.date(LocalDate.now());
                log.debug("REST request to update Like by project id : {}", likes.getProjectID());

                Likes result = likesService.save(like);

                return ResponseEntity.ok().headers(
                        HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, like.getId().toString()))
                        .body(result);
            }
        return ResponseEntity.badRequest().headers(HeaderUtil.createFailureAlert(applicationName,true,ENTITY_NAME,"internalServerError","The projectID was not set")).body(null);

    }

    @DeleteMapping("/unlikeProject/{projectID}")
    public ResponseEntity<Void> deleteLikeWithProjectId(@PathVariable String projectID) {
        log.debug("REST request to delete Likes for project : {}", projectID);
        likesService.deleteByUserIDandProjectID(userService.getUserWithAuthorities().get().getId().intValue(), projectID);;
        return ResponseEntity.noContent()
                .headers(HeaderUtil.createEntityDeletionAlert(applicationName, true, ENTITY_NAME, projectID.toString()))
                .build();
    }

    @GetMapping("/numberOfLikes/{projectID}")
    public ResponseEntity<Integer> getNumberOfLikesForProjectID(@PathVariable String projectID) {
        log.debug("REST request to get number of likes for project {}", projectID);
        Integer n = likesService.findNumberOfLikesByProjectID(projectID);
        return ResponseEntity.ok().body(n);
    }

    @GetMapping("/hasLiked/{projectID}")
    public ResponseEntity<Boolean> hasLikedProject(@PathVariable String projectID) {
        log.debug("REST request to see if user has liked project {}", projectID);
        Boolean bool = false;
        if(userService.getUserWithAuthorities().isPresent()){
        Likes like = likesService.findLikesByUserIDandProjectID(userService.getUserWithAuthorities().get().getId().intValue(), projectID);
        if(like != null){
            log.debug("User has liked this project yet");    
            bool = true;
        }}
        return ResponseEntity.ok().body(bool);
    }
    
    /**
     * {@code GET  /likes} : get all the likes.
     *
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the list
     *         of likes in body.
     */
    @GetMapping("/likes")
    public ResponseEntity<List<Likes>> getAllLikes(LikesCriteria criteria) {
        log.debug("REST request to get Likes by criteria: {}", criteria);
        List<Likes> entityList = likesQueryService.findByCriteria(criteria);
        return ResponseEntity.ok().body(entityList);
    }

    /**
     * {@code GET  /likes/count} : count all the likes.
     *
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the count
     *         in body.
     */
    @GetMapping("/likes/count")
    public ResponseEntity<Long> countLikes(LikesCriteria criteria) {
        log.debug("REST request to count Likes by criteria: {}", criteria);
        return ResponseEntity.ok().body(likesQueryService.countByCriteria(criteria));
    }

    /**
     * {@code GET  /likes/:id} : get the "id" likes.
     *
     * @param id the id of the likes to retrieve.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body
     *         the likes, or with status {@code 404 (Not Found)}.
     */
    @GetMapping("/likes/{id}")
    public ResponseEntity<Likes> getLikes(@PathVariable Long id) {
        log.debug("REST request to get Likes : {}", id);
        Optional<Likes> likes = likesService.findOne(id);
        return ResponseUtil.wrapOrNotFound(likes);
    }

    /**
     * {@code DELETE  /likes/:id} : delete the "id" likes.
     *
     * @param id the id of the likes to delete.
     * @return the {@link ResponseEntity} with status {@code 204 (NO_CONTENT)}.
     */
    @DeleteMapping("/likes/{id}")
    public ResponseEntity<Void> deleteLikes(@PathVariable Long id) {
        log.debug("REST request to delete Likes : {}", id);
        likesService.delete(id);
        return ResponseEntity.noContent()
                .headers(HeaderUtil.createEntityDeletionAlert(applicationName, true, ENTITY_NAME, id.toString()))
                .build();
    }

    /**
     * {@code SEARCH  /_search/likes?query=:query} : search for the likes
     * corresponding to the query.
     *
     * @param query the query of the likes search.
     * @return the result of the search.
     */
    @GetMapping("/_search/likes")
    public List<Likes> searchLikes(@RequestParam String query) {
        log.debug("REST request to search Likes for query {}", query);
        return likesService.search(query);
    }
}
